#include "xio.h"
#include "xbasic_types.h"

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef WORD
#define WORD unsigned short
#endif
#ifndef LWORD
#define LWORD unsigned long
#endif

// Ensure network structures aren't padded (DJGPP and VC)
#pragma pack(1)

#define MACLEN      6           // Ethernet (MAC) address length
#define BCASTADDR 0xff,0xff,0xff,0xff,0xff,0xff // Broadcast address

/* Ethernet Frame sizes */
#define MAXFRAME    1514 // Maximum frame size (inc hdr)
#define MINFRAME    64   // Minimum frame size (inc hdr)
#define ETHERMTU    80   // (MAXFRAME-sizeof(ETHERHDR))

// Ethernet (DIX) header
typedef struct {
  BYTE dest[MACLEN]; // Destination MAC address
  BYTE srce[MACLEN]; // Source MAC address
  WORD ptype;        // Protocol type or length
} ETHERHDR;

// Ethernet frame; data size is frame size minus header & CRC
typedef struct {
  ETHERHDR h;          //Header
  BYTE data[ETHERMTU]; // Data
} ETHERFRAME;

#define WORDMODE 1 // Set to zero if using 8-bit bus

// NE2000 definitions
#define NIC_BASE (0x00A00400) // Base I/O address of the NIC card
#define DATAPORT (0x10*2)
#define NE_RESET (0x1f*2)


// NIC page0 register offsets
#define CMDR    (0x00*2) // command register for read & write
#define PSTART  (0x01*2) // page start register for write
#define PSTOP   (0x02*2) // page stop register for write
#define BNRY    (0x03*2) // boundary reg for rd and wr
#define TPSR    (0x04*2) // tx start page start reg for wr
#define TSR      TPSR
#define TBCR0   (0x05*2) // tx byte count 0 reg for wr
#define TBCR1   (0x06*2) // tx byte count 1 reg for wr
#define ISR     (0x07*2) // interrupt status reg for rd and wr
#define RSAR0   (0x08*2) // low byte of remote start addr
#define RSAR1   (0x09*2) // hi byte of remote start addr
#define CRDA0    RSAR0
#define CRDA1    RSAR1
#define RBCR0   (0x0A*2) // remote byte count reg 0 for wr
#define RBCR1   (0x0B*2) // remote byte count reg 1 for wr
#define RCR     (0x0C*2) // rx configuration reg for wr
#define TCR     (0x0D*2) // tx configuration reg for wr
#define DCR     (0x0E*2) // data configuration reg for wr
#define IMR     (0x0F*2) // interrupt mask reg for wr

// NIC page 1 register offsets
#define PAR0    (0x01*2) // physical addr reg 0 for rd and wr
#define CURR    (0x07*2) // current page reg for rd and wr
#define MAR0    (0x08*2) // multicast addr reg 0 for rd and WR

// Buffer Length and Field Definition Info
#define TXSTART  0x40              // Tx buffer start page
#define TXPAGES  6                 // Pages for Tx buffer
#define RXSTART  (TXSTART+TXPAGES) // Rx buffer start page
#if WORDMODE
#define RXSTOP   0x7e              // Rx buffer end page for word mode
#define DCRVAL   0x49              // DCR values for word mode
#else
#define RXSTOP   0x5f              // Ditto for byte mode
#define DCRVAL   0x48
#endif
#define STARHACK 0                  // Set non-zero to enable Starlan length hack


typedef struct        // Net driver configuration data
{
  BYTE myeth[MACLEN]; // MAC (Ethernet) addr
  WORD next_pkt;      // Next (current) Rx page
} CONFIGNE;

static CONFIGNE nic; // driver condifuration data
int promisc=0;       // Flag to enable promiscuous mode

typedef struct {     // NIC hardware packet header
  BYTE stat;         // Error status
  BYTE next;         // Pointer to next block
  WORD len;          // Length of this frame incl. CRC
} NICHDR;
NICHDR nichdr;

// NE2000-compatible driver prototypes
int init_etherne();

// Private prototypes
void diag();
void resetnic();
WORD innic(int reg);
void outnic(int reg, int b);
void delay(int mult);
void dmaRead(void *data, WORD addr, WORD length);
void dmaWrite(void *data, WORD addr, WORD length);
void send();

WORD packet[1400];

int main(){
  int ok=0;

  print("\r\n\r\n---------------Starting-------------\r\n\r\n");
  diag();
  if(!init_etherne()){
    print("Ethernet NIC not detected");
    return;
  }

  print("\r\n\r\n");
  print("Sending Packets...\r\n");
  while(1){
    send();
  }
  print("\r\n\r\n---------------Done-------------\r\n");
  return 0;
}

// diagnostic that tests nic functionality
void diag(){
  int i;
  
  print("\r\n**************** Beginning of Diagnostic ****************\r\n\r\n");
  
  print("         Command Register Page Switching Test\r\n");
  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);

  print("Writing to and reading from 0x0D (value should be 0x4e): ");
  outnic(0x0D*2,0x4e);
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 0\r\n");
  outnic(CMDR, 0x21);

  print("Reading from reg offset 0x0D: ");
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);
  print("Reading from reg offset 0x0D (should be 0x4e): ");
  putnum(innic(0x0D*2));
  print("\r\n\r\n");

  print("         Default Value Test\r\n");
  print("Switching to page 0\r\n");
  outnic(CMDR,21);
  print("Reading from 0x16 (value should be 0x15): ");
  putnum(innic(0x16*2));
  print("\r\n");
  print("Reading from 0x12 (value should be 0x0c): ");
  putnum(innic(0x12*2));
  print("\r\n");
  print("Reading from 0x13 (value should be 0x12): ");
  putnum(innic(0x13*2));
  print("\r\n\r\n");

  print("         Reset Test\r\n");
  outnic(NE_RESET, innic(NE_RESET));
  if ((innic(ISR) & 0x80) == 0){
    print("  Ethernet card failed to reset!");
  }
  else{
    print("Ethernet card reset successful!");
  }
  print("\r\n\r\n**************** End of Diagnostic ****************\r\n");
}

int init_etherne()
{
  int ok=0;

  outnic(NE_RESET, innic(NE_RESET));  // Do reset
  delay(2);
  if ((innic(ISR) & 0x80) == 0)       // Report if failed
  {
    print("  Ethernet card failed to reset!\r\n");
  }
  else
  {
    print("Ethernet card reset successful\r\n");
    resetnic();                // Reset Ethernet card, get my addr
    ok = 1;
  }
  return(ok);
}

void resetnic(){
  int i;
  WORD eeprom[16] ;

  outnic(CMDR, 0x21);  // Abort and DMA and stop the NIC
  delay(2);

  outnic(DCR, DCRVAL);   // Set word-wide access

  outnic(RBCR0, 0x00); // Clear the count regs
  outnic(RBCR1, 0x00);

  outnic(IMR, 0x00);   // Mask completion irq
  outnic(ISR, 0xFF);   // clear interrupt status register

  outnic(RCR, 0x20);   // 0x20  Set to monitor mode
  //outnic(TCR, 0x02);   // 0x02  and loopback mode

  // Set Rx start, Rx stop and boundary regs
  outnic(PSTART, RXSTART);
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));

  outnic(TPSR, TXSTART);  // Set Tx start page
  
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(IMR, 0x00);   // Mask completion irq

  outnic(CMDR, 0x61);

  // initialize physical address registers (dummy values)
  for(i=0; i<6; i++){
    outnic(PAR0+(i*2), 0xc4);
    print("Initializing Physical Address Register ");
    putnum(i);
    print(" to 0xc4: ");
    putnum(innic(PAR0+(i*2)));
    print("\r\n");
  }

  // initialize CURRent pointer
  outnic(CURR, RXSTART);
  print("\r\n");

  // initialize Multicast Address Registers (accept all)
  for(i=0; i<8; i++){
    outnic(MAR0+(i*2), 0x1e);
    print("Initializing Multicast Address Register ");
    putnum(i);
    print(" to 0x1e: ");
    putnum(innic(MAR0+(i*2)));
    print("\r\n");
  }


  // put nic in start mode
  outnic(CMDR, 0x22); // start nic
  outnic(RCR, 0x44);
  outnic(TCR, 0x00);
  outnic(CMDR, 0x21);
  outnic(DCR, DCRVAL);
  outnic(CMDR, 0x22);
  outnic(ISR, 0xff);
  outnic(IMR, 0x00);
  outnic(TCR, 0x00);  // put nic in normal operation

  /*
    perform dmaRead and dmaWrite
  print("\r\n");
  print("Performing Remote DMA Write of 32-bytes ...\r\n");
  print("(CRDA0, CRDA1, ISR)\r\n");
  dmaWrite(&i, TXSTART<<8, 0x20);
  print("\r\n");
  print("Performing Remote DMA Read of 32-bytes...\r\n");
  print("(DATAPORT, CRDA0, CRDA1, ISR)\r\n");
  dmaRead(&i, TXSTART<<8, 0x20);

  for(i=0; i<0x10000; i+=256){
    dmaRead(&i, i, 0x20);
  }

  for(i=0; i<10; i++){
    outnic(DATAPORT, 0xaaaa);
    putnum(innic(DATAPORT));
    print("\r\n");
  }
  */
}

void dmaRead(void *data, WORD addr, WORD len){
  int i;
  int value;

  // since RBCR goes down to zero, set it equal to len-1
  // when using even number bytes
  // for odd number bytes leave it alone and the last 16-bit
  // read will include the last byte in the upper 8 bits.
  if((len&0x01) == 0){ // even number of bytes
    len--;
  }
  len+=10;
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RBCR0, (len&0xff));
  outnic(RBCR1, (len>>8));
  outnic(RSAR0, 0x3300 | (addr&0xff)); // DMA starting address
  outnic(RSAR1, 0x4400 | (addr>>8));
  outnic(CMDR, 0x08);  // start the DMA read
  delay(2);
  for(i=0; i<=len/2; i++){
    value = innic(DATAPORT);
    putnum(value);
    print(" ");
    putnum(innic(RBCR0)); // display curr DMA address (LO)
    print(" ");
    putnum(innic(RBCR1)); // display curr DMA address (HI)
    print(" ");
    putnum(innic(CRDA0)); // display curr DMA address (LO)
    print(" ");
    putnum(innic(CRDA1)); // display curr DMA address (HI)
    print(" ");
    putnum(innic(ISR)); // interrupt status register (0x40 is DMA complete)
    print("\r\n");
  }
}

void dmaWrite(void *data, WORD addr, WORD len){
  int i;
  int val1, val2, val3, val4;
  // since RBCR goes down to zero, set it equal to len-1
  // when using even number bytes
  // for odd number bytes leave it alone and the last 16-bit
  // write will include the last byte in the upper 8 bits.
  if((len&0x1) == 0){ // even number of bytes
    len--;
  }

  // account for read-before-write problem
  outnic(RBCR0, 0x42);
  outnic(RBCR1, 0x00);
  //outnic(RSAR0, 0x42);
  //outnic(RSAR1, 0x00);
  outnic(CMDR, 0x0a);
  
  //outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RBCR0, (len&0xff));
  outnic(RBCR1, (len>>8));
  outnic(RSAR0, (addr&0xff)); // DMA starting address
  outnic(RSAR1, (addr>>8));
  outnic(CMDR, 0x1e);  // start the DMA write
  delay(2);
  for(i=0; i<=len/2; i++){
    outnic(DATAPORT,0x1234); // write data (hardcoded to 0xee for no reason)
    putnum(innic(CRDA0)); // display curr DMA address (LO)
    print(" ");
    putnum(innic(CRDA1)); // display curr DMA address (HI)
    print(" ");
    putnum(innic(ISR)); // interrupt status register (0x40 is DMA complete)
    print("\r\n");
  }
}

void send(WORD *packet, WORD length){
  register value, count;
  
  while(innic(CMDR)&0x04);

  outnic(TPSR, TXSTART);
  
  // do a dma write

  outnic(TBCR0, 0x20);
  outnic(TBCR1, 0x00);
  outnic(CMDR, 0x26);
  count = 10000;
  putnum(innic(TPSR));
  print("\r\n");
  while((value=innic(TSR)) == 0 && --count);
    
  putnum(count);
  print(" ");
  putnum(value);
  print("\r\n");
}

// Input a word from a NIC register
WORD innic(int reg){
  return(XIo_In16(NIC_BASE+reg));
}

// Output a word to a NIC register
void outnic(int reg, int b){
  XIo_Out16((NIC_BASE+reg), b);
}

void delay(int mult){
  int i;
  int delay = 1000000*mult;
  for(i=0; i<delay; i++);
}
